@@ -13,13 +13,20 @@ package net.i2p.sam;
  *
  * @author human
  */
-public class SAMException extends Exception {
+class SAMException extends Exception {
+
+    private static final long serialVersionUID = 1;
 
     public SAMException() {
-	super();
+    	super();
     }
     
     public SAMException(String s) {
-	super(s);
+    	super(s);
+    }
+    
+    /** @since 0.9.14 */
+    public SAMException(String s, Throwable cause) {
+        super(s, cause);
     }
 }
