@@ -19,10 +19,12 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *  
- * $Revision: 1.1 $
+ * $Revision: 1.2 $
  */
 package i2p.susi.debug;
 
+import net.i2p.I2PAppContext;
+
 /**
  * @author susi23
  */
@@ -31,13 +33,30 @@ public class Debug {
 	public static final int ERROR = 1;
 	public static final int DEBUG = 2;
 	private static int level = ERROR;
+
 	public static void setLevel( int newLevel )
 	{
 		level = newLevel;
-	}	
-	public static void debug( int msgLevel, String msg )
+	}
+
+	/** @since 0.9.13 */
+	public static int getLevel() {
+		return level;
+	}
+
+	public static void debug( int msgLevel, String msg ) {
+		debug(msgLevel, msg, null);
+	}
+
+	/** @since 0.9.34 */
+	public static void debug(int msgLevel, String msg, Throwable t)
 	{
-		if( msgLevel <= level )
-			System.err.println( msg );
+		if( msgLevel <= level ) {
+			System.err.println("SusiMail: " + msg);
+			if (t != null)
+				t.printStackTrace();
+		}
+		if (msgLevel <= ERROR)
+			I2PAppContext.getGlobalContext().logManager().getLog(Debug.class).error(msg, t);
 	}
 }
