@@ -14,8 +14,20 @@ package net.i2p;
  *
  */
 public class CoreVersion {
-    public final static String ID = "$Revision: 1.69 $ $Date: 2006-10-08 20:44:55 $";
-    public final static String VERSION = "0.6.1.26";
+
+    /** deprecated */
+    public final static String ID = "Monotone";
+
+    public final static String VERSION = "0.9.34";
+
+    /**
+     *  For Vuze.
+     *  @return VERSION
+     *  @since 0.9.19
+     */
+    public static String getVersion() {
+        return VERSION;
+    }
 
     public static void main(String args[]) {
         System.out.println("I2P Core version: " + VERSION);
