@@ -3,111 +3,118 @@ package net.i2p.stat;
 import java.io.IOException;
 import java.util.Properties;
 
-import net.i2p.util.Clock;
-import net.i2p.util.Log;
+import net.i2p.data.DataHelper;
 
 /**
  * Simple rate calculator for periodically sampled data points - determining an 
  * average value over a period, the number of events in that period, the maximum number
  * of events (using the interval between events), and lifetime data.
  *
+ * If value is always a constant, you should be using Frequency instead.
  */
 public class Rate {
-    private final static Log _log = new Log(Rate.class);
-    private volatile double _currentTotalValue;
-    private volatile long _currentEventCount;
-    private volatile long _currentTotalEventTime;
-    private volatile double _lastTotalValue;
-    private volatile long _lastEventCount;
-    private volatile long _lastTotalEventTime;
-    private volatile double _extremeTotalValue;
-    private volatile long _extremeEventCount;
-    private volatile long _extremeTotalEventTime;
-    private volatile double _lifetimeTotalValue;
-    private volatile long _lifetimeEventCount;
-    private volatile long _lifetimeTotalEventTime;
+    //private final static Log _log = new Log(Rate.class);
+    private float _currentTotalValue;
+    // was long, save space
+    private int _currentEventCount;
+    private int _currentTotalEventTime;
+    private float _lastTotalValue;
+    // was long, save space
+    private int _lastEventCount;
+    private int _lastTotalEventTime;
+    private float _extremeTotalValue;
+    // was long, save space
+    private int _extremeEventCount;
+    private int _extremeTotalEventTime;
+    private float _lifetimeTotalValue;
+    private long _lifetimeEventCount;
+    private long _lifetimeTotalEventTime;
     private RateSummaryListener _summaryListener;
     private RateStat _stat;
 
-    private volatile long _lastCoalesceDate;
+    private long _lastCoalesceDate;
     private long _creationDate;
-    private long _period;
+    // was long, save space
+    private int _period;
 
     /** locked during coalesce and addData */
-    private Object _lock = new Object();
+    // private final Object _lock = new Object();
 
     /** in the current (partial) period, what is the total value acrued through all events? */
-    public double getCurrentTotalValue() {
+    public synchronized double getCurrentTotalValue() {
         return _currentTotalValue;
     }
 
     /** in the current (partial) period, how many events have occurred? */
-    public long getCurrentEventCount() {
+    public synchronized long getCurrentEventCount() {
         return _currentEventCount;
     }
 
     /** in the current (partial) period, how much of the time has been spent doing the events? */
-    public long getCurrentTotalEventTime() {
+    public synchronized long getCurrentTotalEventTime() {
         return _currentTotalEventTime;
     }
 
     /** in the last full period, what was the total value acrued through all events? */
-    public double getLastTotalValue() {
+    public synchronized double getLastTotalValue() {
         return _lastTotalValue;
     }
 
     /** in the last full period, how many events occurred? */
-    public long getLastEventCount() {
+    public synchronized long getLastEventCount() {
         return _lastEventCount;
     }
 
     /** in the last full period, how much of the time was spent doing the events? */
-    public long getLastTotalEventTime() {
+    public synchronized long getLastTotalEventTime() {
         return _lastTotalEventTime;
     }
 
     /** what was the max total value acrued in any period?  */
-    public double getExtremeTotalValue() {
+    public synchronized double getExtremeTotalValue() {
         return _extremeTotalValue;
     }
 
-    /** when the max(totalValue) was achieved, how many events occurred in that period? */
-    public long getExtremeEventCount() {
+    /**
+     * when the max(totalValue) was achieved, how many events occurred in that period?
+     * Note that this is not necesarily the highest event count; that isn't tracked.
+     */
+    public synchronized long getExtremeEventCount() {
         return _extremeEventCount;
     }
 
     /** when the max(totalValue) was achieved, how much of the time was spent doing the events? */
-    public long getExtremeTotalEventTime() {
+    public synchronized long getExtremeTotalEventTime() {
         return _extremeTotalEventTime;
     }
 
     /** since rate creation, what was the total value acrued through all events?  */
-    public double getLifetimeTotalValue() {
+    public synchronized double getLifetimeTotalValue() {
         return _lifetimeTotalValue;
     }
 
     /** since rate creation, how many events have occurred? */
-    public long getLifetimeEventCount() {
+    public synchronized long getLifetimeEventCount() {
         return _lifetimeEventCount;
     }
 
     /** since rate creation, how much of the time was spent doing the events? */
-    public long getLifetimeTotalEventTime() {
+    public synchronized long getLifetimeTotalEventTime() {
         return _lifetimeTotalEventTime;
     }
 
     /** when was the rate last coalesced? */
-    public long getLastCoalesceDate() {
+    public synchronized long getLastCoalesceDate() {
         return _lastCoalesceDate;
     }
 
     /** when was this rate created? */
-    public long getCreationDate() {
+    public synchronized long getCreationDate() {
         return _creationDate;
     }
 
     /** how large should this rate's cycle be? */
-    public long getPeriod() {
+    public synchronized long getPeriod() {
         return _period;
     }
     
@@ -115,33 +122,23 @@ public class Rate {
     public void setRateStat(RateStat rs) { _stat = rs; }
 
     /**
-     *
-     * @param period number of milliseconds in the period this rate deals with
-     * @throws IllegalArgumentException if the period is not greater than 0
+     * A rate with period shorter than Router.COALESCE_TIME = 50*1000 has to
+     * be manually coalesced before values are fetched from it.
+     * @param period number of milliseconds in the period this rate deals with, min 1, max Integer.MAX_VALUE
+     * @throws IllegalArgumentException if the period is invalid
      */
     public Rate(long period) throws IllegalArgumentException {
-        if (period <= 0) throw new IllegalArgumentException("The period must be strictly positive");
-        _currentTotalValue = 0.0d;
-        _currentEventCount = 0;
-        _currentTotalEventTime = 0;
-        _lastTotalValue = 0.0d;
-        _lastEventCount = 0;
-        _lastTotalEventTime = 0;
-        _extremeTotalValue = 0.0d;
-        _extremeEventCount = 0;
-        _extremeTotalEventTime = 0;
-        _lifetimeTotalValue = 0.0d;
-        _lifetimeEventCount = 0;
-        _lifetimeTotalEventTime = 0;
+        if (period <= 0 || period > Integer.MAX_VALUE)
+            throw new IllegalArgumentException();
 
         _creationDate = now();
         _lastCoalesceDate = _creationDate;
-        _period = period;
+        _period = (int) period;
     }
 
     /**
      * Create a new rate and load its state from the properties, taking data 
-     * from the data points underneath the given prefix.  <p />
+     * from the data points underneath the given prefix.  <p>
      * (e.g. prefix = "profile.dbIntroduction.60m", this will load the associated data points such
      * as "profile.dbIntroduction.60m.lifetimeEventCount").  The data can be exported
      * through store(outputStream, "profile.dbIntroduction.60m").
@@ -157,99 +154,156 @@ public class Rate {
         load(props, prefix, treatAsCurrent);
     }
 
-    /** accrue the data in the current period as an instantaneous event */
-    public void addData(long value) {
-        addData(value, 0);
+    /**
+     * Accrue the data in the current period as an instantaneous event.
+     * If value is always a constant, you should be using Frequency instead.
+     * If you always use this call, eventDuration is always zero,
+     * and the various get*Saturation*() and get*EventTime() methods will return zero.
+     */
+    public synchronized void addData(long value) {
+        _currentTotalValue += value;
+        _currentEventCount++;
+        _lifetimeTotalValue += value;
+        _lifetimeEventCount++;
     }
 
     /**
      * Accrue the data in the current period as if the event took the specified amount of time
+     * If value is always a constant, you should be using Frequency instead.
+     * If eventDuration is nonzero, then the various get*Saturation*() and get*EventTime()
+     * methods will also return nonzero.
+     *
+     * <pre>
+     * There are at least 4 possible strategies for eventDuration:
+     *
+     *   1) eventDuration is always zero.
+     *      The various get*Saturation*() and get*EventTime() methods will return zero.
+     *
+     *   2) Each eventDuration is relatively small, and reflects processing time.
+     *      This is probably the original meaning of "saturation", as it allows you
+     *      to track how much time is spent gathering the stats.
+     *      get*EventTime() will be close to 0.
+     *      get*EventSaturation() will return values close to 0,
+     *      get*SaturationLimit() will return adjusted values for the totals.
+     *
+     *   3) The total of the eventDurations are approximately equal to total elapsed time.
+     *      get*EventTime() will be close to the period.
+     *      get*EventSaturation() will return values close to 1,
+     *      get*SaturationLimit() will return adjusted values for the totals.
+     *
+     *   4) Each eventDuration is not a duration at all, but someother independent data.
+     *      get*EventTime() may be used to retrieve the data.
+     *      get*EventSaturation() are probably useless.
+     *      get*SaturationLimit() are probably useless.
+     * </pre>
      *
      * @param value value to accrue in the current period
      * @param eventDuration how long it took to accrue this data (set to 0 if it was instantaneous)
      */
-    public void addData(long value, long eventDuration) {
-        synchronized (_lock) {
-            _currentTotalValue += value;
-            _currentEventCount++;
-            _currentTotalEventTime += eventDuration;
+    public synchronized void addData(long value, long eventDuration) {
+        _currentTotalValue += value;
+        _currentEventCount++;
+        _currentTotalEventTime += eventDuration;
 
-            _lifetimeTotalValue += value;
-            _lifetimeEventCount++;
-            _lifetimeTotalEventTime += eventDuration;
-        }
+        _lifetimeTotalValue += value;
+        _lifetimeEventCount++;
+        _lifetimeTotalEventTime += eventDuration;
     }
 
     /** 2s is plenty of slack to deal with slow coalescing (across many stats) */
     private static final int SLACK = 2000;
     public void coalesce() {
         long now = now();
-        synchronized (_lock) {
+        double correctedTotalValue; // for summaryListener which divides by rounded EventCount
+        synchronized (this) {
             long measuredPeriod = now - _lastCoalesceDate;
             if (measuredPeriod < _period - SLACK) {
                 // no need to coalesce (assuming we only try to do so once per minute)
-                if (_log.shouldLog(Log.WARN))
-                    _log.warn("not coalescing, measuredPeriod = " + measuredPeriod + " period = " + _period);
+                //if (_log.shouldLog(Log.DEBUG))
+                //    _log.debug("not coalescing, measuredPeriod = " + measuredPeriod + " period = " + _period);
                 return;
             }
     
             // ok ok, lets coalesce
 
             // how much were we off by?  (so that we can sample down the measured values)
-            double periodFactor = measuredPeriod / (double)_period;
+            float periodFactor = measuredPeriod / (float)_period;
             _lastTotalValue = _currentTotalValue / periodFactor;
-            _lastEventCount = (long) ( (_currentEventCount + periodFactor - 1) / periodFactor);
-            _lastTotalEventTime = (long) (_currentTotalEventTime / periodFactor);
+            _lastEventCount = (int) (0.499999 + (_currentEventCount / periodFactor));
+            _lastTotalEventTime = (int) (_currentTotalEventTime / periodFactor);
             _lastCoalesceDate = now;
+            if (_currentEventCount == 0)
+                correctedTotalValue = 0;
+            else
+                correctedTotalValue = _currentTotalValue *
+                                      (_lastEventCount / (double) _currentEventCount);
 
-            if (_lastTotalValue > _extremeTotalValue) {
+            if (_lastTotalValue >= _extremeTotalValue) {  // get the most recent if identical
                 _extremeTotalValue = _lastTotalValue;
                 _extremeEventCount = _lastEventCount;
                 _extremeTotalEventTime = _lastTotalEventTime;
             }
 
-            _currentTotalValue = 0.0D;
+            _currentTotalValue = 0.0f;
             _currentEventCount = 0;
             _currentTotalEventTime = 0;
         }
         if (_summaryListener != null)
-            _summaryListener.add(_lastTotalValue, _lastEventCount, _lastTotalEventTime, _period);
+            _summaryListener.add(correctedTotalValue, _lastEventCount, _lastTotalEventTime, _period);
     }
 
     public void setSummaryListener(RateSummaryListener listener) { _summaryListener = listener; }
     public RateSummaryListener getSummaryListener() { return _summaryListener; }
     
-    /** what was the average value across the events in the last period? */
-    public double getAverageValue() {
-        if ((_lastTotalValue != 0) && (_lastEventCount > 0))
-            return _lastTotalValue / _lastEventCount;
+    /**
+     * What was the average value across the events in the last period?
+     */
+    public synchronized double getAverageValue() {
+        int lec = _lastEventCount;  // avoid race NPE
+        if ((_lastTotalValue != 0) && (lec > 0))
+            return _lastTotalValue / lec;
             
         return 0.0D;
     }
 
-    /** what was the average value across the events in the most active period? */
-    public double getExtremeAverageValue() {
+    /**
+     * During the extreme period (i.e. the period with the highest total value),
+     * what was the average value?
+     */
+    public synchronized double getExtremeAverageValue() {
         if ((_extremeTotalValue != 0) && (_extremeEventCount > 0))
             return _extremeTotalValue / _extremeEventCount;
 
         return 0.0D;
     }
 
-    /** what was the average value across the events since the stat was created? */
-    public double getLifetimeAverageValue() {
+    /**
+     * What was the average value across the events since the stat was created?
+     */
+    public synchronized double getLifetimeAverageValue() {
         if ((_lifetimeTotalValue != 0) && (_lifetimeEventCount > 0))
             return _lifetimeTotalValue / _lifetimeEventCount;
        
         return 0.0D;
     }
 
+    /**
+     * @return the average or lifetime average depending on last event count
+     * @since 0.9.4
+     */
+    public synchronized double getAvgOrLifetimeAvg() {
+        if (getLastEventCount() > 0)
+            return getAverageValue();
+        return getLifetimeAverageValue();
+    }
+    
     /** 
      * During the last period, how much of the time was spent actually processing events in proportion 
      * to how many events could have occurred if there were no intervals?
      *
-     * @return percentage, or 0 if event times aren't used
+     * @return ratio, or 0 if event times aren't used
      */
-    public double getLastEventSaturation() {
+    public synchronized double getLastEventSaturation() {
         if ((_lastEventCount > 0) && (_lastTotalEventTime > 0)) {
             /*double eventTime = (double) _lastTotalEventTime / (double) _lastEventCount;
             double maxEvents = _period / eventTime;
@@ -263,12 +317,13 @@ public class Rate {
     }
 
     /** 
-     * During the extreme period, how much of the time was spent actually processing events
+     * During the extreme period (i.e. the period with the highest total value),
+     * how much of the time was spent actually processing events
      * in proportion to how many events could have occurred if there were no intervals? 
      *
-     * @return percentage, or 0 if the statistic doesn't use event times
+     * @return ratio, or 0 if the statistic doesn't use event times
      */
-    public double getExtremeEventSaturation() {
+    public synchronized double getExtremeEventSaturation() {
         if ((_extremeEventCount > 0) && (_extremeTotalEventTime > 0)) {
             double eventTime = (double) _extremeTotalEventTime / (double) _extremeEventCount;
             double maxEvents = _period / eventTime;
@@ -281,9 +336,9 @@ public class Rate {
      * During the lifetime of this stat, how much of the time was spent actually processing events in proportion 
      * to how many events could have occurred if there were no intervals? 
      *
-     * @return percentage, or 0 if event times aren't used
+     * @return ratio, or 0 if event times aren't used
      */
-    public double getLifetimeEventSaturation() {
+    public synchronized double getLifetimeEventSaturation() {
         if ((_lastEventCount > 0) && (_lifetimeTotalEventTime > 0)) {
             double eventTime = (double) _lifetimeTotalEventTime / (double) _lifetimeEventCount;
             double maxEvents = _period / eventTime;
@@ -295,7 +350,7 @@ public class Rate {
     }
 
     /** how many periods have we already completed? */
-    public long getLifetimePeriods() {
+    public synchronized long getLifetimePeriods() {
         long lifetime = now() - _creationDate;
         double periods = lifetime / (double) _period;
         return (long) Math.floor(periods);
@@ -307,7 +362,7 @@ public class Rate {
      *
      * @return max total value, or 0 if event times aren't used
      */
-    public double getLastSaturationLimit() {
+    public synchronized double getLastSaturationLimit() {
         if ((_lastTotalValue != 0) && (_lastEventCount > 0) && (_lastTotalEventTime > 0)) {
             double saturation = getLastEventSaturation();
             if (saturation != 0.0D) return _lastTotalValue / saturation;
@@ -318,12 +373,13 @@ public class Rate {
     }
 
     /** 
-     * using the extreme period's rate, what is the total value that could have been 
+     * During the extreme period (i.e. the period with the highest total value),
+     * what is the total value that could have been 
      * sent if events were constant?
      *
      * @return event total at saturation, or 0 if no event times are measured
      */
-    public double getExtremeSaturationLimit() {
+    public synchronized double getExtremeSaturationLimit() {
         if ((_extremeTotalValue != 0) && (_extremeEventCount > 0) && (_extremeTotalEventTime > 0)) {
             double saturation = getExtremeEventSaturation();
             if (saturation != 0.0d) return _extremeTotalValue / saturation;
@@ -335,10 +391,11 @@ public class Rate {
     }
 
     /**
-     * How large was the last period's value as compared to the largest period ever?
-     *
+     * What was the total value, compared to the total value in
+     * the extreme period (i.e. the period with the highest total value),
+     * Warning- returns ratio, not percentage (i.e. it is not multiplied by 100 here)
      */
-    public double getPercentageOfExtremeValue() {
+    public synchronized double getPercentageOfExtremeValue() {
         if ((_lastTotalValue != 0) && (_extremeTotalValue != 0))
             return _lastTotalValue / _extremeTotalValue;
         
@@ -347,9 +404,9 @@ public class Rate {
 
     /**
      * How large was the last period's value as compared to the lifetime average value?
-     *
+     * Warning- returns ratio, not percentage (i.e. it is not multiplied by 100 here)
      */
-    public double getPercentageOfLifetimeValue() {
+    public synchronized double getPercentageOfLifetimeValue() {
         if ((_lastTotalValue != 0) && (_lifetimeTotalValue != 0)) {
             double lifetimePeriodValue = _period * (_lifetimeTotalValue / (now() - _creationDate));
             return _lastTotalValue / lifetimePeriodValue;
@@ -357,44 +414,80 @@ public class Rate {
   
         return 0.0D;
     }
+    
+    /**
+     * @return a thread-local temp object containing computed averages.
+     * @since 0.9.4
+     */
+    public RateAverages computeAverages() {
+        return computeAverages(RateAverages.getTemp(),false);
+    }
+    
+    /**
+     * @param out where to store the computed averages.  
+     * @param useLifetime whether the lifetime average should be used if
+     * there are no events.
+     * @return the same RateAverages object for chaining
+     * @since 0.9.4
+     */
+    public synchronized RateAverages computeAverages(RateAverages out, boolean useLifetime) {
+        out.reset();
+        
+        final long total = _currentEventCount + _lastEventCount;
+        out.setTotalEventCount(total);
+        
+        if (total <= 0) {
+            final double avg = useLifetime ? getLifetimeAverageValue() : getAverageValue();
+            out.setAverage(avg);
+        } else {
 
-    public void store(String prefix, StringBuffer buf) throws IOException {
-        PersistenceHelper.add(buf, prefix, ".period", "Number of milliseconds in the period", _period);
-        PersistenceHelper.add(buf, prefix, ".creationDate",
-                              "When was this rate created?  (milliseconds since the epoch, GMT)", _creationDate);
-        PersistenceHelper.add(buf, prefix, ".lastCoalesceDate",
-                              "When did we last coalesce this rate?  (milliseconds since the epoch, GMT)",
+            if (_currentEventCount > 0)
+                out.setCurrent( getCurrentTotalValue() / _currentEventCount );
+            if (_lastEventCount > 0)
+                out.setLast( getLastTotalValue() / _lastEventCount );
+
+            out.setTotalValues(getCurrentTotalValue() + getLastTotalValue());
+            out.setAverage( out.getTotalValues()  / total );
+        }
+        return out;
+    }
+
+    public synchronized void store(String prefix, StringBuilder buf) throws IOException {
+        PersistenceHelper.addTime(buf, prefix, ".period", "Length of the period:", _period);
+        PersistenceHelper.addDate(buf, prefix, ".creationDate",
+                              "When was this rate created?", _creationDate);
+        PersistenceHelper.addDate(buf, prefix, ".lastCoalesceDate",
+                              "When did we last coalesce this rate?",
                               _lastCoalesceDate);
-        PersistenceHelper.add(buf, prefix, ".currentDate",
-                              "When did this data get written?  (milliseconds since the epoch, GMT)", now());
+        PersistenceHelper.addDate(buf, prefix, ".currentDate",
+                              "When was this data written?", now());
         PersistenceHelper.add(buf, prefix, ".currentTotalValue",
                               "Total value of data points in the current (uncoalesced) period", _currentTotalValue);
-        PersistenceHelper
-                         .add(buf, prefix, ".currentEventCount",
+        PersistenceHelper.add(buf, prefix, ".currentEventCount",
                               "How many events have occurred in the current (uncoalesced) period?", _currentEventCount);
-        PersistenceHelper.add(buf, prefix, ".currentTotalEventTime",
-                              "How many milliseconds have the events in the current (uncoalesced) period consumed?",
+        PersistenceHelper.addTime(buf, prefix, ".currentTotalEventTime",
+                              "How much time have the events in the current (uncoalesced) period consumed?",
                               _currentTotalEventTime);
         PersistenceHelper.add(buf, prefix, ".lastTotalValue",
                               "Total value of data points in the most recent (coalesced) period", _lastTotalValue);
         PersistenceHelper.add(buf, prefix, ".lastEventCount",
                               "How many events have occurred in the most recent (coalesced) period?", _lastEventCount);
-        PersistenceHelper.add(buf, prefix, ".lastTotalEventTime",
-                              "How many milliseconds have the events in the most recent (coalesced) period consumed?",
+        PersistenceHelper.addTime(buf, prefix, ".lastTotalEventTime",
+                              "How much time have the events in the most recent (coalesced) period consumed?",
                               _lastTotalEventTime);
         PersistenceHelper.add(buf, prefix, ".extremeTotalValue",
                               "Total value of data points in the most extreme period", _extremeTotalValue);
         PersistenceHelper.add(buf, prefix, ".extremeEventCount",
                               "How many events have occurred in the most extreme period?", _extremeEventCount);
-        PersistenceHelper.add(buf, prefix, ".extremeTotalEventTime",
-                              "How many milliseconds have the events in the most extreme period consumed?",
+        PersistenceHelper.addTime(buf, prefix, ".extremeTotalEventTime",
+                              "How much time have the events in the most extreme period consumed?",
                               _extremeTotalEventTime);
         PersistenceHelper.add(buf, prefix, ".lifetimeTotalValue",
                               "Total value of data points since this stat was created", _lifetimeTotalValue);
         PersistenceHelper.add(buf, prefix, ".lifetimeEventCount",
                               "How many events have occurred since this stat was created?", _lifetimeEventCount);
-        PersistenceHelper.add(buf, prefix, ".lifetimeTotalEventTime",
-                              "How many milliseconds have the events since this stat was created consumed?",
+        PersistenceHelper.addTime(buf, prefix, ".lifetimeTotalEventTime",
+                              "How much total time was consumed by the events since this stat was created?",
                               _lifetimeTotalEventTime);
     }
 
@@ -408,20 +501,20 @@ public class Rate {
      *                       treat the data with as much freshness (or staleness) as appropriate.
      * @throws IllegalArgumentException if the data was formatted incorrectly
      */
-    public void load(Properties props, String prefix, boolean treatAsCurrent) throws IllegalArgumentException {
-        _period = PersistenceHelper.getLong(props, prefix, ".period");
+    public synchronized void load(Properties props, String prefix, boolean treatAsCurrent) throws IllegalArgumentException {
+        _period = PersistenceHelper.getInt(props, prefix, ".period");
         _creationDate = PersistenceHelper.getLong(props, prefix, ".creationDate");
         _lastCoalesceDate = PersistenceHelper.getLong(props, prefix, ".lastCoalesceDate");
-        _currentTotalValue = PersistenceHelper.getDouble(props, prefix, ".currentTotalValue");
-        _currentEventCount = PersistenceHelper.getLong(props, prefix, ".currentEventCount");
-        _currentTotalEventTime = PersistenceHelper.getLong(props, prefix, ".currentTotalEventTime");
-        _lastTotalValue = PersistenceHelper.getDouble(props, prefix, ".lastTotalValue");
-        _lastEventCount = PersistenceHelper.getLong(props, prefix, ".lastEventCount");
-        _lastTotalEventTime = PersistenceHelper.getLong(props, prefix, ".lastTotalEventTime");
-        _extremeTotalValue = PersistenceHelper.getDouble(props, prefix, ".extremeTotalValue");
-        _extremeEventCount = PersistenceHelper.getLong(props, prefix, ".extremeEventCount");
-        _extremeTotalEventTime = PersistenceHelper.getLong(props, prefix, ".extremeTotalEventTime");
-        _lifetimeTotalValue = PersistenceHelper.getDouble(props, prefix, ".lifetimeTotalValue");
+        _currentTotalValue = (float)PersistenceHelper.getDouble(props, prefix, ".currentTotalValue");
+        _currentEventCount = PersistenceHelper.getInt(props, prefix, ".currentEventCount");
+        _currentTotalEventTime = (int)PersistenceHelper.getLong(props, prefix, ".currentTotalEventTime");
+        _lastTotalValue = (float)PersistenceHelper.getDouble(props, prefix, ".lastTotalValue");
+        _lastEventCount = PersistenceHelper.getInt(props, prefix, ".lastEventCount");
+        _lastTotalEventTime = (int)PersistenceHelper.getLong(props, prefix, ".lastTotalEventTime");
+        _extremeTotalValue = (float)PersistenceHelper.getDouble(props, prefix, ".extremeTotalValue");
+        _extremeEventCount = PersistenceHelper.getInt(props, prefix, ".extremeEventCount");
+        _extremeTotalEventTime = (int)PersistenceHelper.getLong(props, prefix, ".extremeTotalEventTime");
+        _lifetimeTotalValue = (float)PersistenceHelper.getDouble(props, prefix, ".lifetimeTotalValue");
         _lifetimeEventCount = PersistenceHelper.getLong(props, prefix, ".lifetimeEventCount");
         _lifetimeTotalEventTime = PersistenceHelper.getLong(props, prefix, ".lifetimeTotalEventTime");
 
@@ -431,23 +524,36 @@ public class Rate {
         coalesce();
     }
 
-    public boolean equals(Object obj) {
-        if ((obj == null) || (obj.getClass() != Rate.class)) return false;
+    /**
+     * This is used in StatSummarizer and SummaryListener.
+     * We base it on the stat we are tracking, not the stored data.
+     */
+    @Override
+    public synchronized boolean equals(Object obj) {
+        if ((obj == null) || !(obj instanceof Rate)) return false;
         if (obj == this) return true;
         Rate r = (Rate) obj;
-        return _period == r.getPeriod() && _creationDate == r.getCreationDate() &&
-        //_lastCoalesceDate == r.getLastCoalesceDate() &&
-               _currentTotalValue == r.getCurrentTotalValue() && _currentEventCount == r.getCurrentEventCount()
-               && _currentTotalEventTime == r.getCurrentTotalEventTime() && _lastTotalValue == r.getLastTotalValue()
-               && _lastEventCount == r.getLastEventCount() && _lastTotalEventTime == r.getLastTotalEventTime()
-               && _extremeTotalValue == r.getExtremeTotalValue() && _extremeEventCount == r.getExtremeEventCount()
-               && _extremeTotalEventTime == r.getExtremeTotalEventTime()
-               && _lifetimeTotalValue == r.getLifetimeTotalValue() && _lifetimeEventCount == r.getLifetimeEventCount()
-               && _lifetimeTotalEventTime == r.getLifetimeTotalEventTime();
+        if (_period != r.getPeriod() || _creationDate != r.getCreationDate())
+            return false;
+        if (_stat == null && r._stat == null)
+            return true;
+        if (_stat != null && r._stat != null) 
+            return _stat.nameGroupDescEquals(r._stat);
+        return false;
     }
 
-    public String toString() {
-        StringBuffer buf = new StringBuffer(2048);
+    /**
+     * It doesn't appear that Rates are ever stored in a Set or Map
+     * (RateStat stores in an array) so let's make this easy.
+     */
+    @Override
+    public synchronized int hashCode() {
+        return DataHelper.hashCode(_stat) ^ _period ^ ((int) _creationDate);
+    }
+
+    @Override
+    public synchronized String toString() {
+        StringBuilder buf = new StringBuilder(2048);
         buf.append("\n\t total value: ").append(getLastTotalValue());
         buf.append("\n\t highest total value: ").append(getExtremeTotalValue());
         buf.append("\n\t lifetime total value: ").append(getLifetimeTotalValue());
@@ -476,37 +582,4 @@ public class Rate {
         // skew periodically
         return System.currentTimeMillis(); //Clock.getInstance().now();
     }
-
-    public static void main(String args[]) {
-        Rate rate = new Rate(1000);
-        for (int i = 0; i < 50; i++) {
-            try {
-                Thread.sleep(20);
-            } catch (InterruptedException ie) { // nop
-            }
-            rate.addData(i * 100, 20);
-        }
-        rate.coalesce();
-        StringBuffer buf = new StringBuffer(1024);
-        try {
-            rate.store("rate.test", buf);
-            byte data[] = buf.toString().getBytes();
-            _log.error("Stored rate: size = " + data.length + "\n" + buf.toString());
-
-            Properties props = new Properties();
-            props.load(new java.io.ByteArrayInputStream(data));
-
-            //_log.error("Properties loaded: \n" + props);
-
-            Rate r = new Rate(props, "rate.test", true);
-
-            _log.error("Comparison after store/load: " + r.equals(rate));
-        } catch (Throwable t) {
-            _log.error("b0rk", t);
-        }
-        try {
-            Thread.sleep(5000);
-        } catch (InterruptedException ie) { // nop
-        }
-    }
 }
