@@ -8,7 +8,6 @@ package net.i2p.data.i2np;
  *
  */
 
-import java.io.ByteArrayOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
@@ -19,18 +18,23 @@ import net.i2p.data.DataStructureImpl;
 import net.i2p.data.Hash;
 import net.i2p.data.SessionKey;
 import net.i2p.data.TunnelId;
-import net.i2p.util.Log;
+//import net.i2p.util.Log;
 
 
 /**
- * Contains the delivery instructions
+ * Contains the delivery instructions for garlic cloves.
+ * Generic "delivery instructions" are used both in tunnel messages
+ * and in garlic cloves, with slight differences.
+ * However,
+ * the tunnel message generator TrivialPreprocessor and reader FragmentHandler do not use this class,
+ * the reading and writing is handled inline there.
  *
  * @author jrandom
  */
 public class DeliveryInstructions extends DataStructureImpl {
-    private final static Log _log = new Log(DeliveryInstructions.class);
-    private boolean _encrypted;
-    private SessionKey _encryptionKey;
+    //private final static Log _log = new Log(DeliveryInstructions.class);
+    //private boolean _encrypted;
+    //private SessionKey _encryptionKey;
     private int _deliveryMode;
     public final static int DELIVERY_MODE_LOCAL = 0;
     public final static int DELIVERY_MODE_DESTINATION = 1;
@@ -47,43 +51,126 @@ public class DeliveryInstructions extends DataStructureImpl {
     private final static int FLAG_MODE_ROUTER = 2;
     private final static int FLAG_MODE_TUNNEL = 3;
     
+    /** @deprecated unused */
+    @Deprecated
     private final static long FLAG_ENCRYPTED = 128;
     private final static long FLAG_MODE = 96;
     private final static long FLAG_DELAY = 16;
     
+    /**
+     *  Immutable local instructions, no options
+     *
+     *  @since 0.9.9
+     */
+    public static final DeliveryInstructions LOCAL = new LocalInstructions();
+
+    /**
+     *  Returns immutable local instructions, or new
+     *
+     *  @since 0.9.20
+     */
+    public static DeliveryInstructions create(byte[] data, int offset) throws DataFormatException {
+        if (data[offset] == 0)
+            return LOCAL;
+        DeliveryInstructions rv = new DeliveryInstructions();
+        rv.readBytes(data, offset);
+        return rv;
+    }
+
     public DeliveryInstructions() {
-        setEncrypted(false);
-        setEncryptionKey(null);
-        setDeliveryMode(-1);
-        setDestination(null);
-        setRouter(null);
-        setTunnelId(null);
-        setDelayRequested(false);
-        setDelaySeconds(0);
+        _deliveryMode = -1;
     }
     
-    public boolean getEncrypted() { return _encrypted; }
-    public void setEncrypted(boolean encrypted) { _encrypted = encrypted; }
-    public SessionKey getEncryptionKey() { return _encryptionKey; }
-    public void setEncryptionKey(SessionKey key) { _encryptionKey = key; }
+    /**
+     * For cloves only (not tunnels), default false, unused
+     * @deprecated unused
+     */
+    @Deprecated
+    public boolean getEncrypted() { return /* _encrypted */ false; }
+    
+    /**
+     * For cloves only (not tunnels), default false, unused
+     * @deprecated unused
+     */
+    @Deprecated
+    public void setEncrypted(boolean encrypted) { /* _encrypted = encrypted; */ }
+
+    /**
+     * For cloves only (not tunnels), default null, unused
+     * @deprecated unused
+     */
+    @Deprecated
+    public SessionKey getEncryptionKey() { return /* _encryptionKey */ null; }
+
+    /**
+     * For cloves only (not tunnels), default null, unused
+     * @deprecated unused
+     */
+    @Deprecated
+    public void setEncryptionKey(SessionKey key) { /* _encryptionKey = key; */ }
+
+    /** default -1 */
     public int getDeliveryMode() { return _deliveryMode; }
+
+    /** @param mode 0-3 */
     public void setDeliveryMode(int mode) { _deliveryMode = mode; }
+
+    /** default null */
     public Hash getDestination() { return _destinationHash; }
+
+    /** required for DESTINATION */
     public void setDestination(Hash dest) { _destinationHash = dest; }
+
+    /** default null */
     public Hash getRouter() { return _routerHash; }
+
+    /** required for ROUTER or TUNNEL */
     public void setRouter(Hash router) { _routerHash = router; }
+
+    /** default null */
     public TunnelId getTunnelId() { return _tunnelId; }
+
+    /** required for TUNNEL */
     public void setTunnelId(TunnelId id) { _tunnelId = id; }
+    
+    /**
+     * default false, unused
+     * @deprecated unused
+     */
+    @Deprecated
     public boolean getDelayRequested() { return _delayRequested; }
+    
+    /**
+     * default false, unused
+     * @deprecated unused
+     */
+    @Deprecated
     public void setDelayRequested(boolean req) { _delayRequested = req; }
+    
+    /**
+     * default 0, unused
+     * @deprecated unused
+     */
+    @Deprecated
     public long getDelaySeconds() { return _delaySeconds; }
+    
+    /**
+     * default 0, unused
+     * @deprecated unused
+     */
+    @Deprecated
     public void setDelaySeconds(long seconds) { _delaySeconds = seconds; }
     
+    /**
+     * @deprecated unused
+     */
+    @Deprecated
     public void readBytes(InputStream in) throws DataFormatException, IOException {
         long flags = DataHelper.readLong(in, 1);
-        if (_log.shouldLog(Log.DEBUG))
-            _log.debug("Read flags: " + flags + " mode: " +  flagMode(flags));
+        //if (_log.shouldLog(Log.DEBUG))
+        //    _log.debug("Read flags: " + flags + " mode: " +  flagMode(flags));
         
+     /****
         if (flagEncrypted(flags)) {
             SessionKey k = new SessionKey();
             k.readBytes(in);
@@ -92,24 +179,28 @@ public class DeliveryInstructions extends DataStructureImpl {
         } else {
             setEncrypted(false);
         }
+      ****/
         
         setDeliveryMode(flagMode(flags));
         switch (flagMode(flags)) {
             case FLAG_MODE_LOCAL:
                 break;
             case FLAG_MODE_DESTINATION:
-                Hash destHash = new Hash();
-                destHash.readBytes(in);
+                //Hash destHash = new Hash();
+                //destHash.readBytes(in);
+                Hash destHash = Hash.create(in);
                 setDestination(destHash);
                 break;
             case FLAG_MODE_ROUTER:
-                Hash routerHash = new Hash();
-                routerHash.readBytes(in);
+                //Hash routerHash = new Hash();
+                //routerHash.readBytes(in);
+                Hash routerHash = Hash.create(in);
                 setRouter(routerHash);
                 break;
             case FLAG_MODE_TUNNEL:
-                Hash tunnelRouterHash = new Hash();
-                tunnelRouterHash.readBytes(in);
+                //Hash tunnelRouterHash = new Hash();
+                //tunnelRouterHash.readBytes(in);
+                Hash tunnelRouterHash = Hash.create(in);
                 setRouter(tunnelRouterHash);
                 TunnelId id = new TunnelId();
                 id.readBytes(in);
@@ -128,11 +219,12 @@ public class DeliveryInstructions extends DataStructureImpl {
     
     public int readBytes(byte data[], int offset) throws DataFormatException {
         int cur = offset;
-        long flags = DataHelper.fromLong(data, cur, 1);
+        int flags = data[cur] & 0xff;
         cur++;
-        if (_log.shouldLog(Log.DEBUG))
-            _log.debug("Read flags: " + flags + " mode: " +  flagMode(flags));
+        //if (_log.shouldLog(Log.DEBUG))
+        //    _log.debug("Read flags: " + flags + " mode: " +  flagMode(flags));
         
+     /****
         if (flagEncrypted(flags)) {
             byte kd[] = new byte[SessionKey.KEYSIZE_BYTES];
             System.arraycopy(data, cur, kd, 0, SessionKey.KEYSIZE_BYTES);
@@ -142,28 +234,32 @@ public class DeliveryInstructions extends DataStructureImpl {
         } else {
             setEncrypted(false);
         }
+      ****/
         
         setDeliveryMode(flagMode(flags));
         switch (flagMode(flags)) {
             case FLAG_MODE_LOCAL:
                 break;
             case FLAG_MODE_DESTINATION:
-                byte destHash[] = new byte[Hash.HASH_LENGTH];
-                System.arraycopy(data, cur, destHash, 0, Hash.HASH_LENGTH);
+                //byte destHash[] = new byte[Hash.HASH_LENGTH];
+                //System.arraycopy(data, cur, destHash, 0, Hash.HASH_LENGTH);
+                Hash dh = Hash.create(data, cur);
                 cur += Hash.HASH_LENGTH;
-                setDestination(new Hash(destHash));
+                setDestination(dh);
                 break;
             case FLAG_MODE_ROUTER:
-                byte routerHash[] = new byte[Hash.HASH_LENGTH];
-                System.arraycopy(data, cur, routerHash, 0, Hash.HASH_LENGTH);
+                //byte routerHash[] = new byte[Hash.HASH_LENGTH];
+                //System.arraycopy(data, cur, routerHash, 0, Hash.HASH_LENGTH);
+                Hash rh = Hash.create(data, cur);
                 cur += Hash.HASH_LENGTH;
-                setRouter(new Hash(routerHash));
+                setRouter(rh);
                 break;
             case FLAG_MODE_TUNNEL:
-                byte tunnelRouterHash[] = new byte[Hash.HASH_LENGTH];
-                System.arraycopy(data, cur, tunnelRouterHash, 0, Hash.HASH_LENGTH);
+                //byte tunnelRouterHash[] = new byte[Hash.HASH_LENGTH];
+                //System.arraycopy(data, cur, tunnelRouterHash, 0, Hash.HASH_LENGTH);
+                Hash trh = Hash.create(data, cur);
                 cur += Hash.HASH_LENGTH;
-                setRouter(new Hash(tunnelRouterHash));
+                setRouter(trh);
                 setTunnelId(new TunnelId(DataHelper.fromLong(data, cur, 4)));
                 cur += 4;
                 break;
@@ -181,24 +277,34 @@ public class DeliveryInstructions extends DataStructureImpl {
     }
     
     
-    private boolean flagEncrypted(long flags) {
+    /**
+     * For cloves only (not tunnels), default false, unused
+     * @deprecated unused
+     */
+/****
+    private static boolean flagEncrypted(long flags) {
         return (0 != (flags & FLAG_ENCRYPTED));
     }
+****/
     
-    private int flagMode(long flags) {
+    /** high bits */
+    private static int flagMode(long flags) {
         long v = flags & FLAG_MODE;
         v >>>= 5;
         return (int)v;
     }
     
-    private boolean flagDelay(long flags) {
+    /**  unused */
+    private static boolean flagDelay(long flags) {
         return (0 != (flags & FLAG_DELAY));
     }
     
     private long getFlags() {
         long val = 0L;
+     /****
         if (getEncrypted())
             val = val | FLAG_ENCRYPTED;
+      ****/
         long fmode = 0;
         switch (getDeliveryMode()) {
             case FLAG_MODE_LOCAL:
@@ -216,21 +322,23 @@ public class DeliveryInstructions extends DataStructureImpl {
         val = val | fmode;
         if (getDelayRequested())
             val = val | FLAG_DELAY;
-        if (_log.shouldLog(Log.DEBUG))
-            _log.debug("getFlags() = " + val);
+        //if (_log.shouldLog(Log.DEBUG))
+        //    _log.debug("getFlags() = " + val);
         return val;
     }
     
     private int getAdditionalInfoSize() {
         int additionalSize = 0;
+     /****
         if (getEncrypted()) {
             if (_encryptionKey == null) throw new IllegalStateException("Encryption key is not set");
             additionalSize += SessionKey.KEYSIZE_BYTES;
         }
+      ****/
         switch (getDeliveryMode()) {
             case FLAG_MODE_LOCAL:
-                if (_log.shouldLog(Log.DEBUG))
-                    _log.debug("mode = local");
+                //if (_log.shouldLog(Log.DEBUG))
+                //    _log.debug("mode = local");
                 break;
             case FLAG_MODE_DESTINATION:
                 if (_destinationHash == null) throw new IllegalStateException("Destination hash is not set");
@@ -259,11 +367,15 @@ public class DeliveryInstructions extends DataStructureImpl {
         int offset = 0;
         offset += getAdditionalInfo(rv, offset);
         if (offset != additionalSize)
-            _log.log(Log.CRIT, "wtf, additionalSize = " + additionalSize + ", offset = " + offset);
+            //_log.log(Log.CRIT, "size mismatch, additionalSize = " + additionalSize + ", offset = " + offset);
+            throw new IllegalStateException("size mismatch, additionalSize = " + additionalSize + ", offset = " + offset);
         return rv;
     }
+
     private int getAdditionalInfo(byte rv[], int offset) {
         int origOffset = offset;
+
+      /****
         if (getEncrypted()) {
             if (_encryptionKey == null) throw new IllegalStateException("Encryption key is not set");
             System.arraycopy(_encryptionKey.getData(), 0, rv, offset, SessionKey.KEYSIZE_BYTES);
@@ -274,24 +386,26 @@ public class DeliveryInstructions extends DataStructureImpl {
             if (_log.shouldLog(Log.DEBUG))
                 _log.debug("Is NOT Encrypted");
         }
+      ****/
+
         switch (getDeliveryMode()) {
             case FLAG_MODE_LOCAL:
-                if (_log.shouldLog(Log.DEBUG))
-                    _log.debug("mode = local");
+                //if (_log.shouldLog(Log.DEBUG))
+                //    _log.debug("mode = local");
                 break;
             case FLAG_MODE_DESTINATION:
                 if (_destinationHash == null) throw new IllegalStateException("Destination hash is not set");
                 System.arraycopy(_destinationHash.getData(), 0, rv, offset, Hash.HASH_LENGTH);
                 offset += Hash.HASH_LENGTH;
-                if (_log.shouldLog(Log.DEBUG))
-                    _log.debug("mode = destination, hash = " + _destinationHash);
+                //if (_log.shouldLog(Log.DEBUG))
+                //    _log.debug("mode = destination, hash = " + _destinationHash);
                 break;
             case FLAG_MODE_ROUTER:
                 if (_routerHash == null) throw new IllegalStateException("Router hash is not set");
                 System.arraycopy(_routerHash.getData(), 0, rv, offset, Hash.HASH_LENGTH);
                 offset += Hash.HASH_LENGTH;
-                if (_log.shouldLog(Log.DEBUG))
-                    _log.debug("mode = router, routerHash = " + _routerHash);
+                //if (_log.shouldLog(Log.DEBUG))
+                //    _log.debug("mode = router, routerHash = " + _routerHash);
                 break;
             case FLAG_MODE_TUNNEL:
                 if ( (_routerHash == null) || (_tunnelId == null) ) throw new IllegalStateException("Router hash or tunnel ID is not set");
@@ -299,31 +413,35 @@ public class DeliveryInstructions extends DataStructureImpl {
                 offset += Hash.HASH_LENGTH;
                 DataHelper.toLong(rv, offset, 4, _tunnelId.getTunnelId());
                 offset += 4;
-                if (_log.shouldLog(Log.DEBUG))
-                    _log.debug("mode = tunnel, tunnelId = " + _tunnelId.getTunnelId() 
-                               + ", routerHash = " + _routerHash);
+                //if (_log.shouldLog(Log.DEBUG))
+                //    _log.debug("mode = tunnel, tunnelId = " + _tunnelId.getTunnelId() 
+                //               + ", routerHash = " + _routerHash);
                 break;
         }
         if (getDelayRequested()) {
-            if (_log.shouldLog(Log.DEBUG))
-                _log.debug("delay requested: " + getDelaySeconds());
+            //if (_log.shouldLog(Log.DEBUG))
+            //    _log.debug("delay requested: " + getDelaySeconds());
             DataHelper.toLong(rv, offset, 4, getDelaySeconds());
             offset += 4;
         } else {
-            if (_log.shouldLog(Log.DEBUG))
-                _log.debug("delay NOT requested");
+            //if (_log.shouldLog(Log.DEBUG))
+            //    _log.debug("delay NOT requested");
         }
         return offset - origOffset;
     }
     
+    /**
+     * @deprecated unused
+     */
+    @Deprecated
     public void writeBytes(OutputStream out) throws DataFormatException, IOException {
         if ( (_deliveryMode < 0) || (_deliveryMode > FLAG_MODE_TUNNEL) ) throw new DataFormatException("Invalid data: mode = " + _deliveryMode);
         long flags = getFlags();
-        if (_log.shouldLog(Log.DEBUG))
-            _log.debug("Write flags: " + flags + " mode: " + getDeliveryMode() 
-                       + " =?= " + flagMode(flags));
+        //if (_log.shouldLog(Log.DEBUG))
+        //    _log.debug("Write flags: " + flags + " mode: " + getDeliveryMode() 
+        //               + " =?= " + flagMode(flags));
         byte additionalInfo[] = getAdditionalInfo();
-        DataHelper.writeLong(out, 1, flags);
+        out.write((byte) flags);
         if (additionalInfo != null) {
             out.write(additionalInfo);
             out.flush();
@@ -331,14 +449,14 @@ public class DeliveryInstructions extends DataStructureImpl {
     }
     
     /**
-     * return the number of bytes written to the target
+     * @return the number of bytes written to the target
      */
     public int writeBytes(byte target[], int offset) {
         if ( (_deliveryMode < 0) || (_deliveryMode > FLAG_MODE_TUNNEL) ) throw new IllegalStateException("Invalid data: mode = " + _deliveryMode);
         long flags = getFlags();
-        if (_log.shouldLog(Log.DEBUG))
-            _log.debug("Write flags: " + flags + " mode: " + getDeliveryMode() 
-                       + " =?= " + flagMode(flags));
+        //if (_log.shouldLog(Log.DEBUG))
+        //    _log.debug("Write flags: " + flags + " mode: " + getDeliveryMode() 
+        //               + " =?= " + flagMode(flags));
         int origOffset = offset;
         DataHelper.toLong(target, offset, 1, flags);
         offset++;
@@ -351,6 +469,7 @@ public class DeliveryInstructions extends DataStructureImpl {
                + getAdditionalInfoSize();
     }
     
+    @Override
     public boolean equals(Object obj) {
         if ( (obj == null) || !(obj instanceof DeliveryInstructions))
             return false;
@@ -358,13 +477,14 @@ public class DeliveryInstructions extends DataStructureImpl {
         return (getDelayRequested() == instr.getDelayRequested()) &&
                (getDelaySeconds() == instr.getDelaySeconds()) &&
                (getDeliveryMode() == instr.getDeliveryMode()) &&
-               (getEncrypted() == instr.getEncrypted()) &&
+               //(getEncrypted() == instr.getEncrypted()) &&
                DataHelper.eq(getDestination(), instr.getDestination()) &&
                DataHelper.eq(getEncryptionKey(), instr.getEncryptionKey()) &&
                DataHelper.eq(getRouter(), instr.getRouter()) &&
                DataHelper.eq(getTunnelId(), instr.getTunnelId());
     }
     
+    @Override
     public int hashCode() {
         return (int)getDelaySeconds() +
                     getDeliveryMode() +
@@ -374,8 +494,9 @@ public class DeliveryInstructions extends DataStructureImpl {
                     DataHelper.hashCode(getTunnelId());
     }
     
+    @Override
     public String toString() {
-        StringBuffer buf = new StringBuffer(128);
+        StringBuilder buf = new StringBuilder(128);
         buf.append("[DeliveryInstructions: ");
         buf.append("\n\tDelivery mode: ");
         switch (getDeliveryMode()) {
@@ -395,11 +516,97 @@ public class DeliveryInstructions extends DataStructureImpl {
         buf.append("\n\tDelay requested: ").append(getDelayRequested());
         buf.append("\n\tDelay seconds: ").append(getDelaySeconds());
         buf.append("\n\tDestination: ").append(getDestination());
-        buf.append("\n\tEncrypted: ").append(getEncrypted());
+        //buf.append("\n\tEncrypted: ").append(getEncrypted());
         buf.append("\n\tEncryption key: ").append(getEncryptionKey());
         buf.append("\n\tRouter: ").append(getRouter());
         buf.append("\n\tTunnelId: ").append(getTunnelId());
         
         return buf.toString();
     }
+
+    /**
+     *  An immutable local delivery instructions with no options
+     *  for efficiency.
+     *
+     *  @since 0.9.9
+     */
+    private static final class LocalInstructions extends DeliveryInstructions {
+        //private static final byte flag = DELIVERY_MODE_LOCAL << 5;  // 0
+
+        @Override
+        public void setEncrypted(boolean encrypted) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void setEncryptionKey(SessionKey key) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public int getDeliveryMode() { return DELIVERY_MODE_LOCAL; }
+
+        @Override
+        public void setDeliveryMode(int mode) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void setDestination(Hash dest) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void setRouter(Hash router) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void setTunnelId(TunnelId id) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void setDelayRequested(boolean req) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void setDelaySeconds(long seconds) {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void readBytes(InputStream in) throws DataFormatException, IOException {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public int readBytes(byte data[], int offset) throws DataFormatException {
+            throw new RuntimeException("immutable");
+        }
+
+        @Override
+        public void writeBytes(OutputStream out) throws DataFormatException, IOException {
+            out.write((byte) 0);
+        }
+
+        @Override
+        public int writeBytes(byte target[], int offset) {
+            target[offset] = 0;
+            return 1;
+        }
+
+        @Override
+        public int getSize() {
+            return 1;
+        }
+
+        @Override
+        public String toString() {
+            return "[DeliveryInstructions: " +
+                   "\n\tDelivery mode: " +
+                   "local]";
+        }
+    }
 }
