@@ -10,16 +10,15 @@ package net.i2p.router;
 
 import java.io.Writer;
 import java.util.ArrayList;
+import java.util.Date;
 import java.util.List;
+
 import net.i2p.data.Hash;
-import net.i2p.data.RouterIdentity;
-import net.i2p.data.i2np.DataMessage;
+import net.i2p.data.router.RouterIdentity;
+import net.i2p.data.i2np.DatabaseLookupMessage;
+import net.i2p.data.i2np.DatabaseSearchReplyMessage;
 import net.i2p.data.i2np.DeliveryStatusMessage;
 import net.i2p.data.i2np.I2NPMessage;
-import net.i2p.data.i2np.DatabaseSearchReplyMessage;
-import net.i2p.data.i2np.DatabaseLookupMessage;
-import net.i2p.data.i2np.TunnelCreateMessage;
-import net.i2p.data.i2np.TunnelCreateStatusMessage;
 import net.i2p.data.i2np.TunnelDataMessage;
 import net.i2p.data.i2np.TunnelGatewayMessage;
 import net.i2p.util.I2PThread;
@@ -30,19 +29,26 @@ import net.i2p.util.Log;
  * Network communication system when it receives messages, and various jobs 
  * periodically retrieve them for processing.
  *
+ * Actually, this doesn't 'pool' anything, since DISPATCH_DIRECT = true.
  */
 public class InNetMessagePool implements Service {
-    private Log _log;
-    private RouterContext _context;
-    private HandlerJobBuilder _handlerJobBuilders[];
-    private List _pendingDataMessages;
-    private List _pendingDataMessagesFrom;
-    private List _pendingGatewayMessages;
+    private final Log _log;
+    private final RouterContext _context;
+    private final HandlerJobBuilder _handlerJobBuilders[];
+
+    /** following 5 unused unless DISPATCH_DIRECT == false */
+    private final List<I2NPMessage> _pendingDataMessages;
+    private final List<Hash> _pendingDataMessagesFrom;
+    private final List<I2NPMessage> _pendingGatewayMessages;
     private SharedShortCircuitDataJob _shortCircuitDataJob;
     private SharedShortCircuitGatewayJob _shortCircuitGatewayJob;
+
     private boolean _alive;
     private boolean _dispatchThreaded;
     
+    /** Make this >= the max I2NP message type number (currently 24) */
+    private static final int MAX_I2NP_MESSAGE_TYPE = 31;
+
     /**
      * If set to true, we will have two additional threads - one for dispatching
      * tunnel data messages, and another for dispatching tunnel gateway messages.
@@ -63,59 +69,80 @@ public class InNetMessagePool implements Service {
     
     public InNetMessagePool(RouterContext context) {
         _context = context;
-        _handlerJobBuilders = new HandlerJobBuilder[32];
-        _pendingDataMessages = new ArrayList(16);
-        _pendingDataMessagesFrom = new ArrayList(16);
-        _pendingGatewayMessages = new ArrayList(16);
-        _shortCircuitDataJob = new SharedShortCircuitDataJob(context);
-        _shortCircuitGatewayJob = new SharedShortCircuitGatewayJob(context);
+        _handlerJobBuilders = new HandlerJobBuilder[MAX_I2NP_MESSAGE_TYPE + 1];
+        if (DISPATCH_DIRECT) {
+            // keep the compiler happy since they are final
+            _pendingDataMessages = null;
+            _pendingDataMessagesFrom = null;
+            _pendingGatewayMessages = null;
+        } else {
+            _pendingDataMessages = new ArrayList<I2NPMessage>(16);
+            _pendingDataMessagesFrom = new ArrayList<Hash>(16);
+            _pendingGatewayMessages = new ArrayList<I2NPMessage>(16);
+            _shortCircuitDataJob = new SharedShortCircuitDataJob(context);
+            _shortCircuitGatewayJob = new SharedShortCircuitGatewayJob(context);
+        }
         _log = _context.logManager().getLog(InNetMessagePool.class);
-        _alive = false;
-        _context.statManager().createRateStat("inNetPool.dropped", "How often do we drop a message", "InNetPool", new long[] { 60*1000l, 60*60*1000l, 24*60*60*1000l });
-        _context.statManager().createRateStat("inNetPool.droppedDeliveryStatusDelay", "How long after a delivery status message is created do we receive it back again (for messages that are too slow to be handled)", "InNetPool", new long[] { 60*1000l, 60*60*1000l, 24*60*60*1000l });
-        _context.statManager().createRateStat("inNetPool.duplicate", "How often do we receive a duplicate message", "InNetPool", new long[] { 60*1000l, 60*60*1000l, 24*60*60*1000l });
-        _context.statManager().createRateStat("inNetPool.droppedTunnelCreateStatusMessage", "How often we drop a slow-to-arrive tunnel request response", "InNetPool", new long[] { 60*60*1000l, 24*60*60*1000l });
-        _context.statManager().createRateStat("inNetPool.droppedDbLookupResponseMessage", "How often we drop a slow-to-arrive db search response", "InNetPool", new long[] { 60*60*1000l, 24*60*60*1000l });
-        _context.statManager().createRateStat("pool.dispatchDataTime", "How long a tunnel dispatch takes", "Tunnels", new long[] { 10*60*1000l, 60*60*1000l, 24*60*60*1000l });
-        _context.statManager().createRateStat("pool.dispatchGatewayTime", "How long a tunnel gateway dispatch takes", "Tunnels", new long[] { 10*60*1000l, 60*60*1000l, 24*60*60*1000l });
+        _context.statManager().createRateStat("inNetPool.dropped", "How often do we drop a message", "InNetPool", new long[] { 60*60*1000l });
+        _context.statManager().createRateStat("inNetPool.droppedDeliveryStatusDelay", "How long after a delivery status message is created do we receive it back again (for messages that are too slow to be handled)", "InNetPool", new long[] { 60*60*1000l });
+        _context.statManager().createRateStat("inNetPool.duplicate", "How often do we receive a duplicate message", "InNetPool", new long[] { 60*60*1000l });
+        //_context.statManager().createRateStat("inNetPool.droppedTunnelCreateStatusMessage", "How often we drop a slow-to-arrive tunnel request response", "InNetPool", new long[] { 60*60*1000l, 24*60*60*1000l });
+        _context.statManager().createRateStat("inNetPool.droppedDbLookupResponseMessage", "How often we drop a slow-to-arrive db search response", "InNetPool", new long[] { 60*60*1000l });
     }
   
-    public HandlerJobBuilder registerHandlerJobBuilder(int i2npMessageType, HandlerJobBuilder builder) {
+    /**
+     * @return previous builder for this message type, or null
+     * @throws ArrayIndexOutOfBoundsException if i2npMessageType is greater than MAX_I2NP_MESSAGE_TYPE
+     */
+    public synchronized HandlerJobBuilder registerHandlerJobBuilder(int i2npMessageType, HandlerJobBuilder builder) {
         HandlerJobBuilder old = _handlerJobBuilders[i2npMessageType];
         _handlerJobBuilders[i2npMessageType] = builder;
         return old;
     }
   
-    public HandlerJobBuilder unregisterHandlerJobBuilder(int i2npMessageType) {
+    /**
+     * @return previous builder for this message type, or null
+     * @throws ArrayIndexOutOfBoundsException if i2npMessageType is greater than MAX_I2NP_MESSAGE_TYPE
+     * @deprecated unused
+     */
+    @Deprecated
+    public synchronized HandlerJobBuilder unregisterHandlerJobBuilder(int i2npMessageType) {
         HandlerJobBuilder old = _handlerJobBuilders[i2npMessageType];
         _handlerJobBuilders[i2npMessageType] = null;
         return old;
     }
     
     /**
-     * Add a new message to the pool, returning the number of messages in the 
-     * pool so that the comm system can throttle inbound messages.  If there is 
+     * Add a new message to the pool.
+     * If there is 
      * a HandlerJobBuilder for the inbound message type, the message is loaded
      * into a job created by that builder and queued up for processing instead
      * (though if the builder doesn't create a job, it is added to the pool)
      *
+     * @param messageBody non-null
+     * @param fromRouter may be null
+     * @param fromRouterHash may be null, calculated from fromRouter if null
+     *
+     * @return -1 for some types of errors but not all; 0 otherwise
+     *         (was queue length, long ago)
      */
     public int add(I2NPMessage messageBody, RouterIdentity fromRouter, Hash fromRouterHash) {
         long exp = messageBody.getMessageExpiration();
         
         if (_log.shouldLog(Log.INFO))
-                _log.info("Received inbound " 
-                          + " with id " + messageBody.getUniqueId()
-                          + " expiring on " + exp
-                          + " of type " + messageBody.getClass().getName());
+                _log.info("Rcvd" 
+                          + " ID " + messageBody.getUniqueId()
+                          + " exp. " + new Date(exp)
+                          + " type " + messageBody.getClass().getSimpleName());
         
         //if (messageBody instanceof DataMessage) {
         //    _context.statManager().getStatLog().addData(fromRouterHash.toBase64().substring(0,6), "udp.floodDataReceived", 1, 0);
         //    return 0;
         //}
         
+        int type = messageBody.getType();
         String invalidReason = null;
-        if (messageBody instanceof TunnelDataMessage) {
+        if (type == TunnelDataMessage.MESSAGE_TYPE) {
             // the IV validator is sufficient for dup detection on tunnel messages, so
             // just validate the expiration
             invalidReason = _context.messageValidator().validateMessage(exp);
@@ -125,17 +152,18 @@ public class InNetMessagePool implements Service {
         
         if (invalidReason != null) {
             int level = Log.WARN;
-            if (messageBody instanceof TunnelCreateMessage)
-                level = Log.INFO;
+            //if (messageBody instanceof TunnelCreateMessage)
+            //    level = Log.INFO;
             if (_log.shouldLog(level))
-                _log.log(level, "Duplicate message received [" + messageBody.getUniqueId() 
-                          + " expiring on " + exp + "]: " + messageBody.getClass().getName() + ": " + invalidReason 
+                _log.log(level, "Dropping message [" + messageBody.getUniqueId() 
+                          + " expiring on " + exp + "]: " + messageBody.getClass().getSimpleName() + ": " + invalidReason 
                           + ": " + messageBody);
-            _context.statManager().addRateData("inNetPool.dropped", 1, 0);
-            _context.statManager().addRateData("inNetPool.duplicate", 1, 0);
+            _context.statManager().addRateData("inNetPool.dropped", 1);
+            // FIXME not necessarily a duplicate, could be expired too long ago / too far in future
+            _context.statManager().addRateData("inNetPool.duplicate", 1);
             _context.messageHistory().droppedOtherMessage(messageBody, (fromRouter != null ? fromRouter.calculateHash() : fromRouterHash));
             _context.messageHistory().messageProcessingError(messageBody.getUniqueId(), 
-                                                                messageBody.getClass().getName(), 
+                                                                messageBody.getClass().getSimpleName(), 
                                                                 "Duplicate/expired");
             return -1;
         } else {
@@ -145,22 +173,22 @@ public class InNetMessagePool implements Service {
         }
 
         boolean jobFound = false;
-        int type = messageBody.getType();
         boolean allowMatches = true;
         
-        if (messageBody instanceof TunnelGatewayMessage) {
+        if (type == TunnelGatewayMessage.MESSAGE_TYPE) {
             shortCircuitTunnelGateway(messageBody);
             allowMatches = false;
-        } else if (messageBody instanceof TunnelDataMessage) {
+        } else if (type == TunnelDataMessage.MESSAGE_TYPE) {
             shortCircuitTunnelData(messageBody, fromRouterHash);
             allowMatches = false;
         } else {
+            // why don't we allow type 0? There used to be a message of type 0 long ago...
             if ( (type > 0) && (type < _handlerJobBuilders.length) ) {
                 HandlerJobBuilder builder = _handlerJobBuilders[type];
 
                 if (_log.shouldLog(Log.DEBUG))
-                    _log.debug("Add message to the inNetMessage pool - builder: " + builder 
-                               + " message class: " + messageBody.getClass().getName());
+                    _log.debug("Add msg to the pool - builder: " + builder 
+                               + " type: " + messageBody.getClass().getSimpleName());
 
                 if (builder != null) {
                     Job job = builder.createJob(messageBody, fromRouter, 
@@ -186,29 +214,33 @@ public class InNetMessagePool implements Service {
                     // was not handled via HandlerJobBuilder
                     _context.messageHistory().droppedOtherMessage(messageBody, (fromRouter != null ? fromRouter.calculateHash() : fromRouterHash));
                     if (type == DeliveryStatusMessage.MESSAGE_TYPE) {
-                        long timeSinceSent = _context.clock().now() - 
-                                            ((DeliveryStatusMessage)messageBody).getArrival();
-                        if (_log.shouldLog(Log.WARN))
-                            _log.warn("Dropping unhandled delivery status message created " + timeSinceSent + "ms ago: " + messageBody);
-                        _context.statManager().addRateData("inNetPool.droppedDeliveryStatusDelay", timeSinceSent, timeSinceSent);
-                    } else if (type == TunnelCreateStatusMessage.MESSAGE_TYPE) {
-                        if (_log.shouldLog(Log.INFO))
-                            _log.info("Dropping slow tunnel create request response: " + messageBody);
-                        _context.statManager().addRateData("inNetPool.droppedTunnelCreateStatusMessage", 1, 0);
+                        // Avoid logging side effect from a horrible UDP EstablishmentManager hack
+                        // We could set up a separate stat for it but don't bother for now
+                        long arr = ((DeliveryStatusMessage)messageBody).getArrival();
+                        if (arr > 10) {
+                            long timeSinceSent = _context.clock().now() - arr;
+                            if (_log.shouldLog(Log.WARN))
+                                _log.warn("Dropping unhandled delivery status message created " + timeSinceSent + "ms ago: " + messageBody);
+                            _context.statManager().addRateData("inNetPool.droppedDeliveryStatusDelay", timeSinceSent);
+                        }
+                    //} else if (type == TunnelCreateStatusMessage.MESSAGE_TYPE) {
+                    //    if (_log.shouldLog(Log.INFO))
+                    //        _log.info("Dropping slow tunnel create request response: " + messageBody);
+                    //    _context.statManager().addRateData("inNetPool.droppedTunnelCreateStatusMessage", 1, 0);
                     } else if (type == DatabaseSearchReplyMessage.MESSAGE_TYPE) {
                         if (_log.shouldLog(Log.INFO))
                             _log.info("Dropping slow db lookup response: " + messageBody);
-                        _context.statManager().addRateData("inNetPool.droppedDbLookupResponseMessage", 1, 0);
+                        _context.statManager().addRateData("inNetPool.droppedDbLookupResponseMessage", 1);
                     } else if (type == DatabaseLookupMessage.MESSAGE_TYPE) {
                         if (_log.shouldLog(Log.DEBUG))
                             _log.debug("Dropping netDb lookup due to throttling");
                     } else {
                         if (_log.shouldLog(Log.WARN))
                             _log.warn("Message expiring on " 
-                                      + (messageBody != null ? (messageBody.getMessageExpiration()+"") : " [unknown]")
+                                      + messageBody.getMessageExpiration()
                                       + " was not handled by a HandlerJobBuilder - DROPPING: " + messageBody, 
                                       new Exception("f00!"));
-                        _context.statManager().addRateData("inNetPool.dropped", 1, 0);
+                        _context.statManager().addRateData("inNetPool.dropped", 1);
                     }
                 } else {
                     String mtype = messageBody.getClass().getName();
@@ -228,17 +260,18 @@ public class InNetMessagePool implements Service {
     }
     
     public int handleReplies(I2NPMessage messageBody) {
-        List origMessages = _context.messageRegistry().getOriginalMessages(messageBody);
-        if (_log.shouldLog(Log.DEBUG))
-            _log.debug("Original messages for inbound message: " + origMessages.size());
-        if (origMessages.size() > 1) {
-            if (_log.shouldLog(Log.DEBUG))
-                _log.debug("Orig: " + origMessages + " \nthe above are replies for: " + messageBody, 
-                           new Exception("Multiple matches"));
+        List<OutNetMessage> origMessages = _context.messageRegistry().getOriginalMessages(messageBody);
+        int sz = origMessages.size();
+        if (sz <= 0)
+            return 0;
+        if (_log.shouldLog(Log.DEBUG)) {
+            _log.debug("Original messages for inbound message: " + sz);
+            if (sz > 1)
+                _log.debug("Orig: " + origMessages + " \nthe above are replies for: " + messageBody);
         }
 
-        for (int i = 0; i < origMessages.size(); i++) {
-            OutNetMessage omsg = (OutNetMessage)origMessages.get(i);
+        for (int i = 0; i < sz; i++) {
+            OutNetMessage omsg = origMessages.get(i);
             ReplyJob job = omsg.getOnReplyJob();
             if (_log.shouldLog(Log.DEBUG))
                 _log.debug("Original message [" + i + "] " + omsg.getReplySelector() 
@@ -249,7 +282,7 @@ public class InNetMessagePool implements Service {
                 _context.jobQueue().addJob(job);
             }
         }
-        return origMessages.size();
+        return sz;
     }
     
     // the following short circuits the tunnel dispatching - i'm not sure whether
@@ -268,13 +301,11 @@ public class InNetMessagePool implements Service {
                 _context.jobQueue().addJob(_shortCircuitGatewayJob);
         }
     }
+
     private void doShortCircuitTunnelGateway(I2NPMessage messageBody) {
         if (_log.shouldLog(Log.DEBUG))
             _log.debug("Shortcut dispatch tunnelGateway message " + messageBody);
-        long before = _context.clock().now();
         _context.tunnelDispatcher().dispatch((TunnelGatewayMessage)messageBody);
-        long dispatchTime = _context.clock().now() - before;
-        _context.statManager().addRateData("tunnel.dispatchGatewayTime", dispatchTime, dispatchTime);
     }
     
     private void shortCircuitTunnelData(I2NPMessage messageBody, Hash from) {
@@ -298,28 +329,37 @@ public class InNetMessagePool implements Service {
     }
     
     public void renderStatusHTML(Writer out) {}
-    public void restart() { 
+
+    /** does nothing since we aren't threaded */
+    public synchronized void restart() { 
         shutdown(); 
         try { Thread.sleep(100); } catch (InterruptedException ie) {}
         startup(); 
     }
-    public void shutdown() {
+
+    /** does nothing since we aren't threaded */
+    public synchronized void shutdown() {
         _alive = false;
-        synchronized (_pendingDataMessages) {
-            _pendingDataMessages.clear();
-            _pendingDataMessagesFrom.clear();
-            _pendingDataMessages.notifyAll();
+        if (!DISPATCH_DIRECT) {
+            synchronized (_pendingDataMessages) {
+                _pendingDataMessages.clear();
+                _pendingDataMessagesFrom.clear();
+                _pendingDataMessages.notifyAll();
+            }
         }
     }
     
-    public void startup() {
+    /** does nothing since we aren't threaded */
+    public synchronized void startup() {
         _alive = true;
         _dispatchThreaded = DEFAULT_DISPATCH_THREADED;
         String threadedStr = _context.getProperty(PROP_DISPATCH_THREADED);
         if (threadedStr != null) {
-            _dispatchThreaded = Boolean.valueOf(threadedStr).booleanValue();
+            _dispatchThreaded = Boolean.parseBoolean(threadedStr);
         }
         if (_dispatchThreaded) {
+            _context.statManager().createRateStat("pool.dispatchDataTime", "How long a tunnel dispatch takes", "Tunnels", new long[] { 10*60*1000l, 60*60*1000l, 24*60*60*1000l });
+            _context.statManager().createRateStat("pool.dispatchGatewayTime", "How long a tunnel gateway dispatch takes", "Tunnels", new long[] { 10*60*1000l, 60*60*1000l, 24*60*60*1000l });
             I2PThread data = new I2PThread(new TunnelDataDispatcher(), "Tunnel data dispatcher");
             data.setDaemon(true);
             data.start();
@@ -329,6 +369,7 @@ public class InNetMessagePool implements Service {
         }
     }
     
+    /** unused unless DISPATCH_DIRECT == false */
     private class SharedShortCircuitDataJob extends JobImpl {
         public SharedShortCircuitDataJob(RouterContext ctx) {
             super(ctx);
@@ -339,9 +380,9 @@ public class InNetMessagePool implements Service {
             I2NPMessage msg = null;
             Hash from = null;
             synchronized (_pendingDataMessages) {
-                if (_pendingDataMessages.size() > 0) {
-                    msg = (I2NPMessage)_pendingDataMessages.remove(0);
-                    from = (Hash)_pendingDataMessagesFrom.remove(0);
+                if (!_pendingDataMessages.isEmpty()) {
+                    msg = _pendingDataMessages.remove(0);
+                    from = _pendingDataMessagesFrom.remove(0);
                 }
                 remaining = _pendingDataMessages.size();
             }
@@ -351,6 +392,8 @@ public class InNetMessagePool implements Service {
                 getContext().jobQueue().addJob(SharedShortCircuitDataJob.this);
         }
     }
+
+    /** unused unless DISPATCH_DIRECT == false */
     private class SharedShortCircuitGatewayJob extends JobImpl {
         public SharedShortCircuitGatewayJob(RouterContext ctx) {
             super(ctx);
@@ -360,8 +403,8 @@ public class InNetMessagePool implements Service {
             I2NPMessage msg = null;
             int remaining = 0;
             synchronized (_pendingGatewayMessages) {
-                if (_pendingGatewayMessages.size() > 0)
-                    msg = (I2NPMessage)_pendingGatewayMessages.remove(0);
+                if (!_pendingGatewayMessages.isEmpty())
+                    msg = _pendingGatewayMessages.remove(0);
                 remaining = _pendingGatewayMessages.size();
             }
             if (msg != null)
@@ -371,34 +414,37 @@ public class InNetMessagePool implements Service {
         }
     }
     
+    /** unused unless router.dispatchThreaded=true */
     private class TunnelGatewayDispatcher implements Runnable {
         public void run() {
             while (_alive) {
                 I2NPMessage msg = null;
                 try {
                     synchronized (_pendingGatewayMessages) {
-                        if (_pendingGatewayMessages.size() <= 0)
+                        if (_pendingGatewayMessages.isEmpty())
                             _pendingGatewayMessages.wait();
                         else
-                            msg = (I2NPMessage)_pendingGatewayMessages.remove(0);
+                            msg = _pendingGatewayMessages.remove(0);
                     }
                     if (msg != null) {
                         long before = _context.clock().now();
                         doShortCircuitTunnelGateway(msg);
                         long elapsed = _context.clock().now() - before;
-                        _context.statManager().addRateData("pool.dispatchGatewayTime", elapsed, elapsed);
+                        _context.statManager().addRateData("pool.dispatchGatewayTime", elapsed);
                     }
                 } catch (InterruptedException ie) {
                     
                 } catch (OutOfMemoryError oome) {
                     throw oome;
-                } catch (Exception e) {
+                } catch (RuntimeException e) {
                     if (_log.shouldLog(Log.CRIT))
                         _log.log(Log.CRIT, "Error in the tunnel gateway dispatcher", e);
                 }
             }
         }
     }
+
+    /** unused unless router.dispatchThreaded=true */
     private class TunnelDataDispatcher implements Runnable {
         public void run() {
             while (_alive) {
@@ -406,24 +452,24 @@ public class InNetMessagePool implements Service {
                 Hash from = null;
                 try {
                     synchronized (_pendingDataMessages) {
-                        if (_pendingDataMessages.size() <= 0) {
+                        if (_pendingDataMessages.isEmpty()) {
                             _pendingDataMessages.wait();
                         } else {
-                            msg = (I2NPMessage)_pendingDataMessages.remove(0);
-                            from = (Hash)_pendingDataMessagesFrom.remove(0);
+                            msg = _pendingDataMessages.remove(0);
+                            from = _pendingDataMessagesFrom.remove(0);
                         }
                     }
                     if (msg != null) {
                         long before = _context.clock().now();
                         doShortCircuitTunnelData(msg, from);
                         long elapsed = _context.clock().now() - before;
-                        _context.statManager().addRateData("pool.dispatchDataTime", elapsed, elapsed);
+                        _context.statManager().addRateData("pool.dispatchDataTime", elapsed);
                     }
                 } catch (InterruptedException ie) {
                     
                 } catch (OutOfMemoryError oome) {
                     throw oome;
-                } catch (Exception e) {
+                } catch (RuntimeException e) {
                     if (_log.shouldLog(Log.CRIT))
                         _log.log(Log.CRIT, "Error in the tunnel data dispatcher", e);
                 }
