@@ -12,6 +12,7 @@ package net.i2p.router;
 /**
  * Defines an executable task
  *
+ * For use by the router only. Not to be used by applications or plugins.
  */
 public interface Job {
     /**
@@ -19,7 +20,7 @@ public interface Job {
      */
     public String getName();
     /** unique id */
-    public int getJobId();
+    public long getJobId();
     /**
      * Timing criteria for the task
      */
