@@ -8,8 +8,9 @@ package net.i2p.router;
  *
  */
 
-import java.io.Writer;
 import java.util.List;
+import java.util.Set;
+
 import net.i2p.data.Hash;
 
 /**
@@ -25,21 +26,9 @@ public interface PeerManagerFacade extends Service {
      *
      * @return List of Hash objects of the RouterIdentity for matching peers
      */
-    public List selectPeers(PeerSelectionCriteria criteria);
-    public List getPeersByCapability(char capability);
+    public List<Hash> selectPeers(PeerSelectionCriteria criteria);
+    public Set<Hash> getPeersByCapability(char capability);
     public void setCapabilities(Hash peer, String caps);
     public void removeCapabilities(Hash peer);
     public Hash selectRandomByCapability(char capability);
 }
-
-class DummyPeerManagerFacade implements PeerManagerFacade {
-    public void shutdown() {}    
-    public void startup() {}
-    public void restart() {}
-    public void renderStatusHTML(Writer out) { }    
-    public List selectPeers(PeerSelectionCriteria criteria) { return null; }
-    public List getPeersByCapability(char capability) { return null; }
-    public void setCapabilities(Hash peer, String caps) {}
-    public void removeCapabilities(Hash peer) {}
-    public Hash selectRandomByCapability(char capability) { return null; }
-}
