@@ -9,7 +9,7 @@ package net.i2p.router.message;
  */
 
 import net.i2p.data.Hash;
-import net.i2p.data.RouterIdentity;
+import net.i2p.data.router.RouterIdentity;
 import net.i2p.data.i2np.GarlicMessage;
 import net.i2p.data.i2np.I2NPMessage;
 import net.i2p.router.HandlerJobBuilder;
@@ -19,9 +19,14 @@ import net.i2p.router.RouterContext;
 /**
  * HandlerJobBuilder to build jobs to handle GarlicMessages
  *
+ * This is the handler for garlic message not received down a tunnel, which is the
+ * case for floodfills receiving netdb messages.
+ * It is not the handler for garlic messages received down a tunnel,
+ * as InNetMessagePool short circuits tunnel messages,
+ * and those garlic messages are handled in InboundMessageDistributor.
  */
 public class GarlicMessageHandler implements HandlerJobBuilder {
-    private RouterContext _context;
+    private final RouterContext _context;
     
     public GarlicMessageHandler(RouterContext context) {
         _context = context;
